; devramx.asm ( ASSEMBLER SOURCE CODE MODULE )
; faked device file for use with truss.asm
;
; Hooks some of the interrupt 21H services - reading/writing and other
; handling of open files - in order to treat a portion of ram as a file
; The ram in question is mapped to the label ram: at the end of this
; module. If XMS is available then extended memory is used instead.
;
; Access to the ramfile must be triggered by an actual file-system reference
; in order to obtain a valid handle.  
;
; This file should be assembled and linked entirely by itself and then
; converted to a memory image with exe2bin. If the result is called 
; devramx.bin, then truss devramx.bin [args ...] will install a tsr with
; device.bin attached to it, and then run the installation code of this
; module before exiting. 

.MODEL tiny

_TEXT SEGMENT PUBLIC 'CODE'
ASSUME cs:_TEXT,ds:_TEXT
FRAME_BYTES equ 16384 ; size of ramfile that is in conventional memory - as
HOOK equ 21H      ; interrupt to listen for
		      ; opposed to XMS. It is the default size unless more is
		      ; from main. This should match DEFAULT_PARS * 16
		      ; This is 2^14
LO_FRAME_MASK equ 3FFFH ; This is 2^14 - 1
HI_FRAME_MASK equ 0C000H ; This is 2^16 - 2^14 

INCLUDE include.asm  ; this defines configurable equates

org 00h         ; this code called after having been copied to the end
                ; of the resident truss code at offset 400H = 40H par 
		; Entry point is at offset 0, unlike .coms, which start
		; at 100h

jmp os_cionn    ; These two need to stay unchanged. In part,
org 03h         ; they constitute a 3 byte "magic number"
jmp _unhook     ; that is checked by the loader in main.c  
os_cionn:
jmp initialize_device
;;;;;;;;;;;;;;;
; device data ;
;;;;;;;;;;;;;;;

; These need to be right here so that main can load them up 
device_names db 16*MAX_FILES dup (0)           ; space for mount point names
termz  db (0)                                  ; mark end of name array
nfiles db (1)

; A per-file structure that
; defines parameters of the ram files
; we manage. This is analogous to a directory entry

MY_FILE struc
used db (0)                   ; = 1 if slot is used
opened db (0)                 ; = 1 if file is open
written db (0)                ; = 1 if file is dirty
file_ptr dd (00000000H)       ; file position  
max_file_ptr dd ?             ; end of file memory
end_of_file dd (00000000H)    ; file position after last write
dos_handle dw (0FFFH)         ; msdos handle for the mount point
file_access db (02H)          ; file access, rw is assumed unless changed by open
file_time dw ?
file_date dw ?
file_size dd ?
xms_handle dw (0)             ; handle of xms memory where file is stored
attributes dw (0)             ; at present not changeable
MY_FILE ends

 
fff MY_FILE MAX_FILES dup (<>) ; MAX_FILES set from makefile.  
nopen db (0)                   ; the number of open ram files
in_use db (0)                  ; the of entrys in the array fff that are used
old_handler_seg dw ?           ; address of old (dos) 21H handler
old_handler_off dw ?
tsr_psp_seg dw ?               ; we don't seem to use this, although it is set
caller_psp_seg dw ?            ; in initialize routine
little_buf db 'xxxxxxxx.xxx'   ; a little scratch buffer
quiet_flag db 0                ; flag set via main and initialize
iii dw (0)                     ; global storage for fff index currently in use
name_flag db (1)               ; flag, 0 = -m option in use. Set in call to 
			       ; initialize
access_req db ?                ; global variables used by open routine
file_name_seg dw ?
file_name_off dw ?
XMSSVR_OFF dw ?                ; address of XMS server, if present
XMSSVR_SEG dw ?                ; other XMS-related parameters
_XMSversion dw ?
_HIMEMversion dw ?
_xms_avail dw ?
_tot_xms  dw ?
_xms_mem_req dw 400H      ; 1024k = 1 meg by default, can be set by user
_xms_mem_per_file dw ?
bytes_per_file dd ?
max_file_index dw ?       ; index of first slot to be used by dups 
using_xms db 00H          ; flag
my_greeting db 'ramfile (2.0) Chicken Haven Software',0DH,0AH,'$'
exit_msg db 'ramfile: unhooking interrupt vector',0DH,0AH,'$'
error_msg db 'There is something wrong',0DH,0AH,'$'
NO_XMSSERVER_MSG db 'No XMS server detected',0DH,0AH,'$'
NO_XMS_HANDLE_MSG db ' =Unable to obtain XMS handle ',0DH,0AH,'$'
MEM_REQ_MSG db ' request: ','$'
MEM_AVAIL_MSG db 'XMS mem (KB) avail: ','$'
MEM_TOTAL_MSG db ' tot: ','$'
MEM_PER_FILE_MSG db ' per file: ','$'
NEW_LINE db 0AH,0DH,'$'

_unhook:                  ; restore original handler for HOOK
			  ; Called from uninst resident truss code
	push ax           ; Store registers used by unhook
	push si 
	push cx
	push ds
	push dx
	mov ax,cs         ; fake out ds to reference seg address of start
	add ax,40H	  ; of this module. When unhook runs, cs = psp seg 
	mov ds,ax         ; of truss code. Start of this module at offset 40H
			  ; Now we don't need segment overrides on data

	; release xms memory if it is being used
	mov al,using_xms
	cmp al,0
	je no_xms_to_free
	mov al,nfiles     ; load cx with number of ramfiles for free loop
	cbw               ; extend by ah = 0
	mov cx,ax
	xor si,si         ; first ram file index
free_loop:
	mov ah,0AH       ; free extended memory block
	mov dx,word ptr fff[si].xms_handle
	call DWORD PTR XMSSVR_OFF 
	add si, SIZE MY_FILE  ; next ram file index
	loop free_loop

no_xms_to_free:
	mov ax, old_handler_off  ; prepare to restore dos int 21h handler
	mov dx,ax
	mov ax, old_handler_seg
	mov ds,ax
	mov al, HOOK
	mov ah, 25H        ; set interrupt vector
	int 21H	
	pop dx             ; restore registers
	pop ds
	pop cx
	pop si
	pop ax
	ret

error_handler:
	mov dx, offset error_msg
	mov ah, 09H          ; Display greeting
	int 21H
	jmp go_home

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Initialize device: called from   ;	
; truss code after it is installed ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; far *initialize_device(unsigned char, unsigned char, unsigned) 

initialize_device:            ; Program entry point label 
; Initialization code: Hook interrupt 21H services to create ramfiles
	
	push bp           ; set up stack frame
	mov bp,sp

			  ; preserve registers changed in this call
	                  ; using existing stack when we are called
	push bx
	push cx
	push es
	push dx
	push ds
	push si

	mov ax,cs	  ; adjust ds to point to start 
	add ax,40H        ; of this module, which is at offset 40H in tsr
	mov ds,ax

	mov ax,cs         ; seg address of tsr psp. Currently not used
	mov word ptr tsr_psp_seg,ax

	; Danger, Will Robinson. Accessing arguments passed by C run time
        ; is tricky, because it is impossible to predict with certainty
        ; how the compiler will do it. Borland C promotes chars to ints
	; when they are passed.
	mov ax,[bp + 10]       ; access unsigned arg, 
	mov word ptr _xms_mem_req,ax  ; in k bytes  
	mov al,[bp + 8]         ; access 2nd char arg 
	mov byte ptr quiet_flag,al ;
	mov al,[bp + 6]         ; access unsigned char arg (passed as int)
	mov byte ptr name_flag,al;

	; initialize file structures 
	mov cx,MAX_FILES
	mov si,00H
initialize_loop:
	mov byte ptr fff[si].opened, 00H
	mov byte ptr fff[si].written, 00H
	mov word ptr fff[si].end_of_file, 0000H
	mov word ptr fff[si].end_of_file[2], 0000H
	mov word ptr fff[si].xms_handle, 0FFFFH
	mov word ptr fff[si].dos_handle, 00FFH
	mov byte ptr fff[si].used, 00H
	mov word ptr fff[si].attributes, 0000H
	add si, SIZE MY_FILE
loop	initialize_loop

; get and store address of current handler

	mov al, HOOK
	mov ah, 35H       ; get interrupt vector
	int 21H
	mov ax,es
	mov old_handler_seg,ax
	mov old_handler_off,bx

; check to see if an XMS server V2+ installed

	mov ax,4300H    ; Int 2F service XMS Installation Check
	int 2FH
	cmp al,80H      ; Succeeds if service available
	je  xms_ok      
	mov dx, FRAME_BYTES
	mov WORD PTR fff[0].max_file_ptr, dx 
	xor dx,dx
	mov WORD PTR fff[0].max_file_ptr[2],dx
	mov byte ptr fff[0].used,01H
	inc in_use 
	mov dx,offset NO_XMSSERVER_MSG
	mov ah,09H
	int 21H
	jmp go_home


xms_ok:
	; Record that we are using xms so server address is known to be valid
	mov BYTE PTR using_xms, 01H
	; Get the server address and store in XMSSVR_SEG and XMSSVR_OFF
	mov ax, 4310H
	int 2FH        ; ES:BX now has server address
	mov ax,ES
	mov XMSSVR_SEG, ax
	mov XMSSVR_OFF,bx

	; The following gives the paradigm for calling on the server

	; Get available memory
	mov ah, 08H     ; Query free extended memory
	mov bl, 00h     ; XMS function returns error code in bl
	call DWORD PTR XMSSVR_OFF 

	mov _xms_avail,ax
	mov _tot_xms,dx

	; Get the version numbers
	xor ah,ah
	call DWORD PTR XMSSVR_OFF
	mov _XMSversion,ax
	mov _HIMEMversion,bx
	call print_version

; Figure out how much xms memory to give each file
; Adjust request based on total available

	mov ax, _xms_avail
	cmp _xms_mem_req, ax
	jb no_xms_adj
	mov _xms_mem_req, ax

no_xms_adj:
	
; Divide xms memory request by number of files
	mov al,nfiles
	cbw 
	mov cx,ax
	xor dx,dx
	mov ax, _xms_mem_req
	div cx
	mov word ptr _xms_mem_per_file, ax
	mov dx, 400H        ; convert to bytes 
	mul dx              
	mov word ptr bytes_per_file,ax
	mov word ptr bytes_per_file[2],dx

	cmp quiet_flag,0H 
	jne skip_report

	; report on memory usage and availability
	
	mov dx, offset MEM_AVAIL_MSG
	mov ah,09H
	int 21H
	mov ax,_xms_avail
	call print_num

	mov dx,offset MEM_TOTAL_MSG
	mov ah,09H
	int 21H
	mov ax,_tot_xms
	call print_num

	mov dx, offset MEM_REQ_MSG
	mov ah, 09H
	int 21H
	mov ax, word ptr _xms_mem_req
	call print_num

	mov dx, offset MEM_PER_FILE_MSG
	mov ah, 09H
	int 21H
	mov ax, word ptr _xms_mem_per_file
	call print_num
	mov dx,offset NEW_LINE
	mov ah,09H
	int 21H

skip_report:

	mov ax, word ptr _xms_mem_req
	cmp ax,0H
	je no_xms_mem_req

; loop over nfiles giving each one a handle and setting
; its max_file_ptr. 

	xor si,si  ; zero out si, cx still contains nfiles  

handle_loop:	

; Get an XMS handle
	mov ah,09H        ; Allocate Extended Memory Block
	mov dx,_xms_mem_per_file
	call DWORD PTR XMSSVR_OFF
	cmp ax,0
	je  got_no_xms_handle 
	mov word ptr fff[si].xms_handle,dx
	mov ax,word ptr bytes_per_file
	mov word ptr fff[si].max_file_ptr,ax
	mov ax,word ptr bytes_per_file[2]
	mov word ptr fff[si].max_file_ptr[2],ax
	mov byte ptr fff[si].used,01H
	inc in_use
	add si, SIZE MY_FILE
	loop handle_loop
	xor ax,ax
	mov al,nfiles
	mov dl, SIZE MY_FILE
	mul dl
	mov max_file_index,ax
	jmp go_home

; BUG: if this fails partway through then there will be un-freed xms

got_no_xms_handle:
	mov byte ptr NO_XMS_HANDLE_MSG, bl
	mov dx, offset NO_XMS_HANDLE_MSG 
	mov ah,09H                       
	int 21H
no_xms_mem_req:
	mov dx, FRAME_BYTES
	mov WORD PTR fff[0].max_file_ptr, dx 
	xor dx,dx
	mov WORD PTR fff[0].max_file_ptr[2],dx
	mov BYTE PTR using_xms, 00H
	mov byte ptr nfiles, 01H
	mov word ptr max_file_index, SIZE MY_FILE 
	
go_home:                   

; point vector to us
	mov ax, offset new_handler
	mov dx,ax
	mov al, HOOK
	mov ah, 25H        ; set interrupt vector
	int 21H	

; return to startup truss code leaving this stuff resident
	pop si
	pop ds             
	pop dx
	pop es
	pop cx
	pop bx
	pop bp 

	retf               ; far return to truss code 	

new_handler:               ; Get here on int21H
			   ; Unlike initialization code, when we come here
	                   ; cs points to the beginning of this module
			   ; and ds points who knows where?

	pushf              ; save flag values on entry since cmp's change them


; Here we handle services we always intercept, whether there is
; an open ramfile or not. These are passed a file name rather
; than a handle, so they all use the file_name_search routine
 
        cmp ah,3DH  ; open file with handle
	jne other_file_name_services	
	call case_3DH
	jc error_return

good_return:

	         ;  handle is returned in ax 
	popf     ; restore flags 
	retf 2   ; hard lesson: don't iret here because it will
                 ; popf and obliterate any error return

error_return:
	; error is returned in ax 
	popf ; restore flags 
	stc  ; but set carry to indicate error
	retf 2

other_file_name_services:
	cmp ax,4300H  ; get file attributes
	jne aaaa
	call case_4300H
	jc error_return
	jmp good_return

aaaa:
	cmp ax,6CH    ; extended open/create
	jne bbbb
	call case_6CH
	jc error_return
	jmp good_return

bbbb:
	cmp ax,3CH    ; create file with handle
	jne cccc
	call case_3CH
	jc error_return
	jmp good_return

cccc:
	cmp ax,5BH    ; create new file
	jne dddd
	call case_5BH
	jc error_return
	jmp good_return

dddd:
	cmp ax,4EH    ; find first file
	jne eeee
	call case_4EH
	jc error_return
	jmp good_return
	
; here we handle other (non-handle) services 
eeee:
; none at present


; Next we handle only services that apply when some ramfile is open

	cmp byte ptr cs:nopen, 00H
	je orig_vector  ; get out quickly when no ramfile is open so as not
	                ; to unduly bog down int 21H     

	; We intercept 4CH (End Program) in order to "close" our
	; open ramfiles 
	cmp ah,4CH
 	jne xxx
	;mov byte ptr cs:nopen, 00H
	push cx
	push si 
	mov cx,MAX_FILES
	mov si,0

close_all:
	mov al, byte ptr cs:fff[si].opened
	cmp al, 00H
	je next_close_all ; don't close a file that is not open
	mov ax, si
	mov cs:iii,ax     ; store index of file in global iii for use by 
                          ; close routine, case_3EH
 	call case_3EH     ; close ramfile virtually and also close handle
                          ; to mount point in dos. Closing any open non
                          ; ramfiles is left to the actual dos 4CH	
next_close_all:
	add si, SIZE MY_FILE  ; next ramfile slot
	loop close_all
	mov byte ptr cs:nopen, 00H  ; this should not be needed
	pop si
	pop cx
	popf
	mov ah, 4CH ; this will not recurse because nopen = 0
	int 21H     ; does not return 

xxx:

	; In all other cases we only take action on services
	; that supply one of our own handles in bx
	; The handle search is a drag on the system, so it is
	; good for ramfile to allow few files open at a time

	push ax
	push bx
	push cx
	mov cx,MAX_FILES
	mov ax,bx
	mov bx,0

handle_search:
	cmp byte ptr cs:fff[bx].opened,00H
	je next_handle
	cmp word ptr cs:fff[bx].dos_handle,ax ; see mov above	
	je found_handle
next_handle:
	add bx, SIZE MY_FILE
loop handle_search
	pop cx
	pop bx            ; restore handle
	pop ax

; Handle not found. Go to original vector and don't return here
orig_vector:
	popf                 ; restore flags on entry
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf


;    bx now has index of file with desired handle
;    store it for future reference in iii (global variable)

found_handle:
	mov word ptr cs:iii,bx  ; store found file index in global iii
	pop cx                  ; restore, to end handle search cycle
	pop bx  
	pop ax

	; switch to service selected by ah - ones we handle, anyway

handle_service_switch:

	cmp ah,3EH
 	jne eee
	call case_3EH
	jmp common_iret

eee:
	cmp ah,3FH
	jne bbb
	call case_3FH
	jmp common_iret
bbb:
	cmp ah,40H
	jne ccc
	call case_40H
	jmp common_iret
ccc:
	cmp ah,42H
	jne ddd 
	call case_42H
	jmp common_iret
ddd: 
	cmp ax,4400H
	jne qqq
	call case_4400H
	jmp common_iret
qqq:  
	cmp ax,440AH
	jne ggg
	call case_4400H  ; for us, this is the same as 4400H
	jmp common_iret
ggg:
	cmp ax,4406H
	jne ppp
	call case_4406H
	jmp common_iret
ppp:
	cmp ah,45H
	jne rrr
	call case_45H
	jmp common_iret
rrr:
	cmp ah,46H
	jne sss
	call case_46H
	jmp common_iret
sss: 
	jmp orig_vector   ; not one of the services we care about  
			  ; so pass control to handler

common_iret:
	jnc normal_iret
	; carry indicates error condition. 
	; ERROR_INVALID_HANDLE can never occur for us
	popf
	stc
	retf 2
	
normal_iret:
	popf  ; restore flags pushed at start  
	clc
	retf 2

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; File system services    ;;
;; Called when ramfiles    ;;
;; closed, or when file    ;;
;; name rather than handle ;;
;; is passed               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

case_6CH proc

; Extended open/create (DOS 4+). Basically, we just pass this off
; to ordinary open (3DH) if this is for a ramfile. If the 2's bit
; of dx is set, then we truncate the ramfile in question to zero length

	push bx
	push dx
	push ax
	mov dx,si
	call file_name_search
	cmp ax,0
	jne orig_6CH

	pop ax
	call case_3DH	
	jc bad_open6ch 
	mov cx,01H     ; action opened
	pop dx
	cmp dx, 10H
	jne no_create6ch
	mov word ptr cs:fff[bx].end_of_file,00H
	mov word ptr cs:fff[bx].end_of_file[2],00H
	mov cx, 02H    ; say that we created the file if we were asked to
	jmp notrunc6ch  

no_create6ch:
	and dx,02H     ; see if we are supposed to zero out file
	cmp dx,00H
	je notrunc6ch
	mov word ptr cs:fff[bx].end_of_file,00H
	mov word ptr cs:fff[bx].end_of_file[2],00H
	mov cx,03H
	jmp notrunc6ch
bad_open6ch:
	pop dx
notrunc6ch:
	pop bx
	ret

orig_6CH:
	pop dx
	pop bx
	pop ax

; Now we call the original int 21H handler, but return here

	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again6ch
	push ax
	mov ah,6CH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf      

back_again6ch:
	ret
endp
case_3DH proc

; Open file with handle 
; On entry ds:dx points to file name and al has
; access request
; returns filesystem handle in ax if successful, carry set and error code
; in ax otherwise  
	

; save stuff 
	mov byte ptr cs:access_req,al  ; save for later use
	mov word ptr cs:file_name_off,dx 
	mov ax,ds
	mov word ptr cs:file_name_seg,ax
	mov cs:iii,0000H               ; default index of mount point

	push ds
	push es
	push dx
	push bx
	push cx

; see if we are going to go by our stored mount point name(s) (-m option)
; or to allow any read-only zero size file as mount point

	mov al,byte ptr cs:name_flag
	cmp al,00H
	je read_only_zero_size

	call file_name_search
	mov cs:iii,bx  ; save index of found file
	cmp ax,0000H
	je ok_to_open2 ; file found, so open it

	jmp open_with_prefs ; search failed, so open with original access req

read_only_zero_size:

; Now we call the original int 21H handler, but return here
; This same code is used 4 different places, but in-lined for speed

	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again
	push ax
	mov al,00H ; open read only for testing
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again:
	jc open_with_prefs  ; try to open with original access request 

	mov bx,ax    ; handle returned from open in bx

	; See if it is a device rather than a file
	; If it were, e.g., CON, the read test for zero length below
        ; would hang waiting for input, so we don't want to do that
	; 

	mov ax,4400H ; Get device data
	int 21H	
	cmp dx,128
	jb its_a_file    ; it's a device if bit 7 is set (= 128)
	mov ah,3EH       ; close device
	int 21H
	jmp open_with_prefs

ok_to_open2:
jmp ok_to_open           ; dratted 128 byte limit!

its_a_file:

	; else, it's a file
	; try to read one byte from the open file. If it succeeds,
	; iret to whomever wanted the handle -
        ; not us, since we only deal with place-holder files of zero length

	mov cx,0001H ; ask to read one byte
	mov ax,cs    ; load ds for this segment
	mov ds,ax
	mov dx, offset little_buf   ; ds:dx = our little buffer
	mov ah,3FH     ; read file or device
	int 21H
	cmp ax,0000H   ; see how many bytes were read
	je zero_size_file  ; if none, then we are good to go so far

	mov ah,3EH ; close file
	int 21H
	jmp open_with_prefs
	
zero_size_file:

	; close the file and try to open for write. If this fails
	; then we are definitely good to go - file is read-only
	mov ah,3EH   ; close file
	int 21H

	mov ax,word ptr cs:file_name_seg
	mov ds,ax
	mov dx,word ptr cs:file_name_off
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again2
	push ax
	mov al,02H ; rw for testing
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again2:
	jc ok_to_open

	mov ah,3EH   ; close file
	int 21H

open_with_prefs:

	mov ax,word ptr cs:file_name_seg
	mov ds,ax
	mov dx,word ptr cs:file_name_off
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again4
	push ax
	mov al,byte ptr cs:access_req ; stored original parameters
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again4:
	jmp  open_restore_regs
	
ok_to_open:
	; open place-holder file for good
	mov ax,word ptr cs:file_name_seg
	mov ds,ax
	mov dx,word ptr cs:file_name_off
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_again3
	push ax
	mov al,00H ; open place-holder file read-only
	mov ah,3DH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; wheeeeeeee....!

back_again3:
	mov bx,cs:iii
	mov word ptr cs:fff[bx].dos_handle,ax      ; store handle
	mov byte ptr cs:fff[bx].opened,01H          ; mark the ramfile as open
	mov word ptr cs:fff[bx].file_ptr[2], 0000H  ; zero-out file pointer
	mov word ptr cs:fff[bx].file_ptr, 0000H ; 
	mov byte ptr cs:fff[bx].written, 00H        ; zero-out dirty file flag
	inc cs:nopen  ; new open ram file
;	cmp byte ptr cs:fff[bx].used,00H
;	jne open_restore_regs
;	mov byte ptr cs:fff[bx].used,01H
;	inc cs:in_use

open_restore_regs:

	pop cx
	pop bx
	pop dx
	pop es
	pop ds
	ret
endp

;;;;;;;;;;;;;;;;;;;;;;
; Get File Attributes;
;;;;;;;;;;;;;;;;;;;;;;
;
; At present, the only possible attribute for a ramfile
; is ATTR_NORMAL (0000H), so this routine just returns
; that attribute if ds:dx points to a ramfile mount point
; and otherwise hands off to the original dos handler using
; retf so as not to cause recursion
; The attribute is returned in cx. The DOS call may generate
; an error (carry bit set and error code returned in ax)
;
; There is no need at present to intercept the companion
; 4301H Set File Attributes

case_4300H proc

	push bx                   ; file name search uses this register
	call file_name_search     ; see if ds:dx points to our mount point
	cmp ax, 0000H             ; mount point [bx] name matches
	jne old_4300
	mov cx,cs:fff[bx].attributes ; load cx with found file attributes
	jmp ret_4300                 ; clean up and return 

old_4300:
; call original handler
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset ret_4300
	push ax
	mov ax,4300H
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; go to dos
ret_4300:
	pop bx     ; back here again, as arranged
	ret        ; return via service switch
endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; case_3CH create file with
; handle
; If the path given is not a 
; ramfile, call original
; handler. 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

case_3CH proc
	push bx                   ; file name search uses this register
	call file_name_search     ; see if ds:dx points to our mount point
	cmp ax, 0000H             ; mount point [bx] name matches
	jne old_3CH
	mov al, 02H               ; create ramfile means just open rw
	call case_3DH             ; open ram file and truncate to zero length
	jc ret_3CH
	mov word ptr cs:fff[bx].end_of_file,00H
	mov word ptr cs:fff[bx].end_of_file[2],00H
	jmp ret_3CH

old_3CH:
; call original handler
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset ret_3CH
	push ax
	mov ah, 3CH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; go to dos
ret_3CH:
	pop bx     ; back here again, as arranged
	ret        ; return via service switch
endp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; case_5BH create new file with
; handle
; If the path given is not a 
; ramfile, call original
; handler. 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

case_5BH proc

	push bx                   ; file name search uses this register
	call file_name_search     ; see if ds:dx points to our mount point
	cmp ax, 0000H             ; mount point [bx] name matches
	jne old_5BH
	mov al, 02H               ; create ramfile means just open rw
	call case_3DH             ; open ram file and truncate to zero length
	jc bad_return5BH
	mov word ptr cs:fff[bx].end_of_file,00H
	mov word ptr cs:fff[bx].end_of_file[2],00H
	jmp ret_5BH

bad_return5BH:
	jmp ret_5BH	

old_5BH:
; call original handler
	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset ret_5BH
	push ax
	mov ah, 5BH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; go to dos
ret_5BH:
	pop bx     ; back here again, as arranged
	ret        ; return via service switch
endp

FILEINFO STRUC
	fiReserved   db 21 dup (?) ; reserved
	fiAttribute  db ?          ; attributes of file found
	fiFileTime   dw ?          ; time of last write
	fiFileDate   dw ?          ; date of last write
	fiSize	     dd ?          ; file size (bytes)
	fiFileName   db 13 dup (?) ; file name . extension (+ null char)
FILEINFO ENDS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Find First File
; If successful, carry flag is clear and it fills DTA table
; Else carry is set and error value returned in ax
; Ours doctors what dos does, to mark a ramfile as
; always normal independent of mount point attributes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

case_4EH proc

; Search to see if ds:dx points to a mount point
; BUG: we really should use the found filename so we can
; handle wildcard expressions. Bit of a pain because it lacks disk letter

	push bx                   ; file name search uses this register
	call file_name_search     ; see if ds:dx points to our mount point
	cmp ax, 0000H             ; mount point [bx] name matches
	je old_4EH                ; and bx is a valid mount point index
	mov bx, 0FFFH             ; this value signals no mount point found

old_4EH:

; Call the dos interrupt in line to 
; fill a fileinfo structure at the current DTA for 
; the first found file 

	pushf      ; the first three push's get us back here
	push cs    ; after iret from the old handler
	mov ax, offset back_4EH
	push ax
	mov ah, 4EH
	push cs:old_handler_seg ; do far jump to original handler
	push cs:old_handler_off
	retf       ; go to dos

back_4EH:
;  continue only if there is no carry
	jc ret_4EH

	cmp bx,0FFFH ; was a mount point requested?
	je ret_4EH   ; no, done, since dos has already done the right thing
	
	; else we need to doctor the attributes entry 
	; found file name is a ramfile so we mark it as a normal file
	
	mov ah, 2FH   ; Get current DTA in es:bx
	int 21H
	mov byte ptr es:[bx + 21],00H   ; mark file as normal 

ret_4EH:
	pop bx     ;
	ret        ; return via service switch
endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Handle Services - called ;;
;; When ramfile is open     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

case_42H proc
   ; move file pointer
	; cx-dx has hi-lo words of displacement (signed 32 bit)
	; al has method 00 = from start, 01 = from current, 02 = from end
	push bx
	mov bx, word ptr cs:iii
	cmp al,00H
	je from_start
	cmp al,01H
	je from_current
	cmp al,02H
	je from_end
	mov ax, 0001H ; error ERROR_INVALID_FUNCTION
	stc
	pop bx
	ret
from_start:
	mov word ptr cs:fff[bx].file_ptr[2],cx
	mov word ptr cs:fff[bx].file_ptr,dx
	jmp done42
from_current:
	add word ptr cs:fff[bx].file_ptr,dx    ; N.B. move past eof is possible
	adc word ptr cs:fff[bx].file_ptr[2],cx       
	jmp done42
from_end: 
	add dx,word ptr cs:fff[bx].end_of_file
	adc cx,word ptr cs:fff[bx].end_of_file[2]
	mov word ptr cs:fff[bx].file_ptr,dx
	mov word ptr cs:fff[bx].file_ptr[2],cx
done42:
	mov ax, word ptr cs:fff[bx].file_ptr
	mov dx, word ptr cs:fff[bx].file_ptr[2]
	pop bx
	ret
endp

case_3EH  proc 

; close file with handle 
	push bx
	push ax
	push ds
	push cx
	push dx
	push si

	mov ax,cs                     ; load ds to access this segment
	mov ds,ax

	mov bx, word ptr iii          ; fetch file index
	cmp byte ptr fff[bx].opened,00H
	je ok_to_close                ; close dos handle in dos and return
	dec nopen
	mov byte ptr fff[bx].opened,00H       ; mark this ram-file as closed

	; file marked closed, but we reclaim the slot if it was created
	; by a dup or dup2, i.e., if the handle is greater than that of
	; all the real files 

	cmp bx, max_file_index
	jb skip_reclaim
	mov fff[bx].used, 00H
	dec in_use

skip_reclaim:

	cmp byte ptr fff[bx].written,00H   ; has the file been modified?
; BUG: we need to handle timestamp, i.e. update dirent 
	je ok_to_close             ; no, so just go ahead and close

ok_to_close:
	mov ax, fff[bx].dos_handle
	mov word ptr fff[bx].dos_handle,00FFFH ; restore dos_handle to invalid value
				      ; doing here makes sure int 21H
                                      ; calls below do not recurse
	mov cl,nopen
	mov nopen,00H ; prevent recursion
	mov bx,ax
	mov ah,3EH    ; Actually ask Dos to close handle of placeholder file
	int 21H       ; Will not recurse because ramfile now closed
	mov nopen,cl

	pop si
	pop dx
	pop cx
	pop ds
	pop ax
	pop bx
	ret          ; return possibly with carry flag set on error
endp

case_46H proc
	; Force duplicate file handle (dup2) 
	; create a copy of an open handle that
	; refers to the same file as the original. Initially, file
	; ptr and other data is the same for both handles.
	; Unlike 45H (dup), dup2 makes sure to provide a requested
	; new handle value, or an error. If the requested handle is
	; open, it is first closed.
	; On entry bx has open handle and cx has requested new
        ; duplicate handle number; 
	; For a ramfile, referring to the same file means having the
	; same xms_handle
	; On error, carry flag is set and error code is returned in ax

	push bx      ; save registers used
	push cx
	push dx
	push ds
	push si
	push di

	mov ax,cs    ; load ds to access this segment
	mov ds,ax

	; get index (si) of file referenced by handle in bx 
	mov si,word ptr iii
	cmp bx, fff[si].dos_handle
	jne no_can_2dup6  ; 6 is ERROR_INVALID_HANDLE

	; recall cx has requested duplicate
	mov dx,cx     ; make a copy for good luck
        ; now we must see if there is an open file slot in fff array
	; or if the requested handle is already there
	mov cx, MAX_FILES  ; counter
	mov di,0           ; index for found slot
	mov ax,0FFFH       ; ridiculous value for last open slot 
slot_search:
	cmp fff[di].dos_handle,dx
	je found_slot2     ; it doesn't matter if open or not
	cmp fff[di].used, 00H
	jne skip_slot      ; move on
	mov ax,di          ; keep last open slot index in ax
skip_slot:
	add di, SIZE MY_FILE
	loop slot_search
	cmp ax, 0FFFH      ; not equal if open slot was found
	jne found_slot2

no_can_2dup4:
	mov ax, 0004H    ; Error too many open files
	jmp no_can_dup2
no_can_2dup6:
	mov ax, 0006H    ; Error invalid handle
no_can_dup2:
	stc  ; ax may contain error code from DOS 46H call
	jmp dup2_return

found_slot2:  
	mov di,ax
	; di is now index of chosen slot
	; if it is open, it must be a case of request handle
	; already in use as a ramfile, so overall there will be
	; no change in our number of open files
	cmp fff[di].opened, 00H
	je skip_virtual_close
	mov fff[di].opened, 00H  ; mark as closed ram file
	dec nopen                ; and decrease open file count

skip_virtual_close:

	; invoke dos dup2, but to prevent recursion we must
	; temporarily change source handle value
	mov fff[si].dos_handle, 0FFFH   ; invalid handle 
	mov cx,dx     ; restore requested handle number, bx has source
	mov ah, 46H
	int 21H       ; dos force duplicate handle system call	
;call debug
	mov fff[si].dos_handle,bx  ; restore referenced file handle
	jc no_can_dup2 ; carry set indicates dos call error	

	; Since the dos call was successful, it has
	; closed the request handle in dos if it needed it. This close is
        ; part of dos int 46H (or maybe not, maybe dos invokes
        ; its own int 3EH - close handle - from within 46H?). Either way, we 
	; are setting up our own slot anew in the next block
	
	cmp byte ptr fff[di].used, 00H
	jne set_it_up
	inc in_use 

set_it_up:
	mov word ptr fff[di].dos_handle, dx ; requested handle number
	mov byte ptr fff[di].used, 01H      ; continuation with copy
	mov ax, word ptr fff[si].file_ptr   ; of source handle
	mov word ptr fff[di].file_ptr,ax
	mov ax,word ptr fff[si].file_ptr[2]
	mov word ptr fff[di].file_ptr[2],ax
	mov ax, word ptr fff[si].end_of_file
	mov word ptr fff[di].end_of_file,ax
	mov ax,word ptr fff[si].end_of_file[2]
	mov word ptr fff[di].end_of_file[2],ax
	mov ah, byte ptr fff[si].written
	mov byte ptr fff[di].written, ah
	mov fff[di].opened,01H
	mov ax, word ptr fff[si].max_file_ptr
	mov word ptr fff[di].max_file_ptr,ax
	mov ax,word ptr fff[si].max_file_ptr[2]
	mov word ptr fff[di].max_file_ptr[2],ax
	mov byte ptr fff[di].file_access, 02H   ; access is always rw. Bug? 
	mov ax,word ptr fff[si].xms_handle
	mov word ptr fff[di].xms_handle, ax
	clc
	inc nopen

dup2_return:
	pop di        ; restore registers used
	pop si
	pop ds
	pop dx
	pop cx
	pop bx
	ret

case_46H endp

case_45H proc

	; Duplicate file handle (dup): create a copy of an open handle that
	; refers to the same file as the original. Initially, file
	; ptr and other data is the same for both handles.
	; On entry bx has old handle; we return new one in ax
	; For a ramfile, referring to the same file means having the
	; same xms_handle
	; On error, carry flag is set and error code is returned in ax

;call debug
	push bx      ; save registers used
	push cx
	push dx
	push ds
	push si
	push di

	mov ax,cs    ; load ds to access this segment
	mov ds,ax

	mov cl, in_use	
	cmp cl,MAX_FILES 
	jae no_can_dup4  ; 4 is ERROR_TOO_MANY_OPEN_FILES

	; get index (si) of file referenced by handle in fff[iii] 
	mov si,word ptr iii
	cmp bx, fff[si].dos_handle
	jne no_can_dup6  ; 6 is ERROR_INVALID_HANDLE

	; we need to get a duplicate handle from dos
	; for the mount point. To prevent recursion
	; we must temporarily change nopen to 0 

	mov cl,nopen
	mov nopen, 00H
	mov ah, 45H
	int 21H       ; dos duplicate handle system call	
	mov nopen, cl
	mov dx,ax     ; new handle or error saved in dx
	jc no_can_dup ; carry set indicates dos call error	

	; find the first open slot, since now we know there is at least one

	mov cx, MAX_FILES  ; counter
	mov di,0           ; index for found slot

slot_search2:
	cmp fff[di].used, 00H
	je found_slot
	add di, SIZE MY_FILE
	loop slot_search2

no_can_dup4:
	mov ax, 0004H    ; Error too many open files
	jmp no_can_dup
no_can_dup6:
	mov ax, 0006H    ; Error invalid handle
no_can_dup:
	stc  ; ax may contain error code from DOS 45H call
	jmp dup_return

found_slot:
	; di is now index of first not used slot
	; Set it up ...
	mov word ptr fff[di].dos_handle, dx ;
	mov byte ptr fff[di].used, 01H 
	mov ax, word ptr fff[si].file_ptr
	mov word ptr fff[di].file_ptr,ax
	mov ax,word ptr fff[si].file_ptr[2]
	mov word ptr fff[di].file_ptr[2],ax
	mov ax, word ptr fff[si].end_of_file
	mov word ptr fff[di].end_of_file,ax
	mov ax,word ptr fff[si].end_of_file[2]
	mov word ptr fff[di].end_of_file[2],ax
	mov fff[di].opened,01H
	mov fff[di].written,00H
	mov ax, word ptr fff[si].max_file_ptr
	mov word ptr fff[di].max_file_ptr,ax
	mov ax,word ptr fff[si].max_file_ptr[2]
	mov word ptr fff[di].max_file_ptr[2],ax
	mov al,byte ptr fff[si].file_access
	mov byte ptr fff[di].file_access, al
	mov ax,word ptr fff[si].xms_handle
	mov word ptr fff[di].xms_handle, ax
	clc
	inc in_use
	inc nopen
	mov ax,dx     ; return new handle

dup_return:
	pop di        ; restore registers used
	pop si
	pop ds
	pop dx
	pop cx
	pop bx
	ret

case_45H endp

case_4400H proc
	; Get device data
	; or 440A, is file or device remote?  (no, actually)
	push bx
	mov bx, word ptr cs:iii
	cmp byte ptr cs:fff[bx].written,00H
	je not_written
	mov dx,0002H     ; written, and disk = 2 (C:)
	pop bx
	ret
not_written:
	mov dx,0042H     ; not written, and disk = 2
	pop bx
	ret
endp

case_4406H proc
	; Is device ready for input ?
	; or 4407H, for output ?
	mov al,0FFH ; yes, it is
	ret

endp

; Structure used to move blocks to and from extended memory
; Set Handle to 0 when referencing conventional memory
; in which case offset is segment:offset pair, but bass-ackwards: segment
; is low byte and offset is high byte - go figure!

ExtMemMoveStruct struc
 Lngth          dd ? ; 32-bit number of bytes to transfer
 SourceHandle    dw ? ; Handle of source block
 SourceOffset    dd ? ; 32-bit offset into source
 DestHandle      dw  ? ; Handle of destination block
 DestOffset      dd  ? ; 32-bit offset into destination
ExtMemMoveStruct ends

myxms ExtMemMoveStruct <0,,,,,> ; dos system calls are limited to moving 64k
                                ; bytes at a time, so we make sure high byte
                                ; is initialized to 0

; Copy frame buffer to extended memory. Destination frame is the one
; containing current value of file ptr  
; Does no bounds checking or reset of file ptr

write_frame proc

	push ax
	push bx
	push cx
	push dx
	push ds                     ; save registers
	push si
	mov ax,cs
	mov ds,ax                   ; load ds for our data segment
	xor ax,ax
	mov myxms.SourceHandle,ax   ; mark as source in conventional mem 
	mov Word Ptr myxms.SourceOffset[0],offset ram ; here we give
	mov Word PTR myxms.SourceOffset[2],ds         ; seg:offset pair
	mov bx,word ptr iii
	mov ax,word ptr fff[bx].xms_handle   
	mov word ptr myxms.DestHandle,ax   ; handle to our xms mem
	mov ax,word ptr fff[bx].file_ptr[0]
	and ax,HI_FRAME_MASK        ; round to (next below) nearest 
			            ; FRAME_BYTES bytes. Write behind
				    
	mov word ptr myxms.DestOffset[0],ax
	mov ax,word ptr fff[bx].file_ptr[2]
	mov word ptr myxms.DestOffset[2],ax
	mov WORD PTR myxms.Lngth[0],FRAME_BYTES 
	mov Word PTR myxms.Lngth[2],00H
	mov si,offset myxms            ; ds:si -> move struct
	mov ah, 0Bh                    ; Move extended mem block
	call DWORD PTR XMSSVR_OFF      ; via xms server call
	cmp ax,00H
	jnz no_xms_write_error         ; success when ax not zero
	mov byte ptr xms_write_message,bl
	mov dx, offset xms_write_message 
	mov ah, 09H          ; Display greeting
	int 21H
no_xms_write_error:
	pop si                         ; restore registers
	pop ds
	pop dx
	pop cx
	pop bx
	pop ax
	ret
endp

xms_write_message db '  = error writing xms memory',0DH,0AH,'$'

; Bring in frame buffer full from extended memory
; to frame buffer, beginning at current value of file_ptr 
; (which should be congruent to zero modulo FRAME_BYTES)
; Does no bounds checking, no reset of file ptr

load_frame proc

	push ax
	push bx
	push cx
	push dx
	push ds                   ; save registers
	push si
	mov ax,cs                 ; load ds with our data segment
	mov ds,ax
	mov bx,word ptr iii
	xor ax,ax
	mov myxms.DestHandle,ax   ; mark destination as in conventional mem 
	mov Word Ptr myxms.DestOffset[0],offset ram  ; dest is our frame
	mov Word PTR myxms.DestOffset[2],ds          ; buffer
	mov ax,word ptr fff[bx].xms_handle   
	mov word ptr myxms.SourceHandle,ax   ; handle to our xms mem
	; reduce file ptr modulo FRAME_BYTES = bytes in frame buffer
	mov ax,word ptr fff[bx].file_ptr[0] ; lo byte sufficient for div
	and ax,HI_FRAME_MASK        ; ax = file_ptr rounded down 
;	mov ax,word ptr fff[bx].file_ptr[0] ; source is mblock beginning at file ptr
	mov word ptr myxms.SourceOffset[0],ax
	mov ax,word ptr fff[bx].file_ptr[2]
	mov word ptr myxms.SourceOffset[2],ax
	mov WORD PTR myxms.Lngth[0], FRAME_BYTES 
	mov WORD PTR myxms.Lngth[2], 00H
	mov si,offset myxms            ; ds:si -> move struct
	mov ah, 0Bh                    ; Move extended mem block
	call DWORD PTR XMSSVR_OFF      ; by calling xms server
	cmp ax,00H
	jnz no_xms_read_error
	mov byte ptr xms_read_message,bl
	mov dx, offset xms_read_message 
	mov ah, 09H          ; Display greeting
	int 21H
no_xms_read_error:
	pop si                         ; restore registers
	pop ds
	pop dx
	pop cx
	pop bx
	pop ax
	ret
endp

xms_read_message db ' = xms read error',0DH,0AH,'$'

case_3FH proc
   ; read file or device
   ; on entry ds:dx points to caller buffer to read into
   ; and cx has number of bytes to read

	push bx
	mov bx, word ptr cs:iii
	cmp byte ptr cs:fff[bx].file_access,01H
	jne ok_to_read
	mov ax,0005H ; ERROR_ACCESS_DENIED
	stc          ; flag error
	pop bx
	ret

ok_to_read:

	push ds         ; save registers 
	push si
	push es
	push di
	push dx

	mov di,dx
	mov ax,ds
	mov es,ax       ; es:di now points to caller buffer to read into
	mov ax,cs
	mov ds,ax       ; load ds -- point to this data segment

	mov dx, cx      ; save original bytes to copy in dx	
	mov ax, word ptr fff[bx].end_of_file[2];
	sub ax, word ptr fff[bx].file_ptr[2];
	cmp ax,0
	ja return_save  ; there is clearly enough space since dos read requests
                        ; can never exceed 64k
	mov ax, word ptr fff[bx].end_of_file[0]
	sub ax, word ptr fff[bx].file_ptr[0]
	cmp ax,cx
	ja return_save ; there is enough space
	mov dx,ax      ; adjust request if not. On return, this indicates eof	
return_save:
	push dx        ; save for later return
refill_loop:
	; reduce file ptr modulo FRAME_BYTES = bytes in frame buffer
	mov ax,word ptr fff[bx].file_ptr[0] ; lo byte sufficient for div
	and ax,LO_FRAME_MASK        ; ax = file_ptr mod FRAME_BYTES
;	cmp ax,0
;	ja skip_load    ; no need to refill frame: file ptr is not at end 
	cmp using_xms,0
	je skip_load
	call load_frame ; refill frame buffer from xms
skip_load:
	mov cx, FRAME_BYTES
	sub cx, ax      ; cx has bytes in frame from pointer onwards
	cmp cx, dx      ; are there enough bytes remaining in frame for copy?
	jb no_trunc     ; yes, don't adjust number to copy
	mov cx,dx       ; else copy only up to request

no_trunc:
	mov si,offset ram                ; ready to read next frame
	add si,ax                        ; aim si
	mov ax,cx                        ; preserve copy number
	cld
	rep movsb                        ; copy to caller buffer
	add word ptr fff[bx].file_ptr[0],ax         ; advance 32 bit file ptr
	adc word ptr fff[bx].file_ptr[2],0H
	sub dx,ax                        ; decrement request
	cmp dx,0                         ; and quit loop if request filled
	ja refill_loop
	pop ax          ; return in ax actual bytes copied
	clc             ; indicate no error
	pop dx
	pop di
	pop es
	pop si
	pop ds
	pop bx
	ret
endp

case_40H proc

 ; write file or device

	     ; on entry ds:dx points to buffer containing data
	     ; cx has requested bytes to write (if possible)

	push bx
	mov bx, word ptr cs:iii
	cmp byte ptr cs:fff[bx].file_access,00H ; read-only
	jne ok_to_write
	mov ax,0005H ; ERROR_ACCESS_DENIED
	stc
	pop bx
	ret

ok_to_write:

	push ds         ; preserve other registers used
	push dx
	push si
	push es
	push di

	mov si,dx       ; ds:si points to buffer to write from
	mov ax,cs
	mov es,ax

			; note that we are stuck with caller's ds,
			; so segment over-ride must be used on all
			; memory accesses to this (cs) segment 


			; make sure there is room to satisfy 
                        ; request, otherwise shorten request
	cmp cs:using_xms,00H
	je use_frame_eof
	mov ax,400H
	mul word ptr cs:_xms_mem_per_file ; convert ram file in xms size to bytes
	sub dx, word ptr cs:fff[bx].end_of_file[2];
	cmp dx,0
	ja enough_space  ; there is clearly enough space since dos write requests
                        ; can never exceed 64k
	sub ax, word ptr cs:fff[bx].end_of_file[0]
	cmp ax,cx
	ja enough_space ; there is enough space before eof
	mov cx,ax      ; adjust request if not. On return, this indicates eof	
	jmp enough_space

use_frame_eof:
	mov ax,word ptr cs:fff[bx].max_file_ptr[0]
	sub ax, word ptr cs:fff[bx].end_of_file[0]
	cmp ax,cx
	ja enough_space
	mov cx,ax

enough_space:
	mov dx,cx      ; save adjusted request size 
	push dx        ; save for return

refill_loop2:

	; reduce file ptr modulo FRAME_BYTES = bytes in frame buffer
	mov ax,word ptr cs:fff[bx].file_ptr[0] ; lo byte sufficient for div
	and ax,LO_FRAME_MASK           ; ax = file_ptr mod FRAME_BYTES
	mov cx, FRAME_BYTES
	sub cx, ax      ; cx has bytes in frame from pointer onwards
	cmp cx, dx      ; are there enough bytes remaining in frame for copy?
	jb no_trunc2    ; yes, don't adjust number to copy
	mov cx,dx       ; else copy only up to request

no_trunc2:
	mov di,offset ram                ; ready to read next frame
	add di,ax                        ; aim di
	mov ax,cx                        ; preserve copy number
	cld
	rep movsb                        ; copy from caller buffer
   	mov byte ptr cs:fff[bx].written,01H
	sub dx,ax                        ; decrement request
	cmp cs:using_xms,0
	je skip_write
	call write_frame                 ; write-behind update ramfile in xms
skip_write:
	add word ptr cs:fff[bx].file_ptr[0],ax
	adc word ptr cs:fff[bx].file_ptr[2],0H
	mov ax, word ptr cs:fff[bx].file_ptr[0]
	mov word ptr cs:fff[bx].end_of_file[0], ax
	mov ax, word ptr cs:fff[bx].file_ptr[2]
	mov word ptr cs:fff[bx].end_of_file[2],ax
	cmp dx,0                         ; and quit loop if no request left
	ja refill_loop2

	pop ax          ; return number of bytes actually written
	clc             ; signal that no error occurred
	pop di
	pop es
	pop si
	pop dx
	pop ds
	pop bx
	ret
endp
debug_msg db '  (int 21H service) ',0DH,0AH,'$'
debug_hdr db 'AH AL BH BL CH CL DH DL + 20H',0DH,0AH,'$'
debug_spacer db '  ','$'
debug_newline db 0DH,0AH,'$'
debug_ah db ?
debug_al db ?
debug_bh db ?
debug_bl db ?
debug_ch db ?
debug_cl db ?
debug_dh db ?
debug_dl db ?


debug proc

	push ds
	push bx
	push ax
	push dx
	mov cs:debug_ah,ah
	mov cs:debug_al,al
	add cs:debug_al,20H ; make readable by moving into printable range
	mov cs:debug_bh,bh
	add cs:debug_bh,20H 
	mov cs:debug_bl,bl
	add cs:debug_bl,20H 
	mov cs:debug_ch,ch
	add cs:debug_ch,20H 
	mov cs:debug_cl,cl
	add cs:debug_cl,20H 
	mov cs:debug_dh,dh
	add cs:debug_dh,20H 
	mov cs:debug_dl,dl
	add cs:debug_dl,20H 

	mov bx, 0FFFH       ; prevent recurse
	mov ax,cs           ; access this segment
	mov ds,ax

	mov ah,09H
	mov dx, offset debug_newline
	int 21H
	mov ah,06H
	mov dl,debug_ah
	int 21H
	mov dx, offset debug_msg
	mov ah, 09H          ; Display interrupt announcement
	int 21H
	mov dx, offset debug_hdr
	mov ah, 09H          ; Display header
	int 21H

	mov ah,06H
	add debug_ah,20H   ; make readable by moving into printable range
	mov dl,debug_ah
	int 21H
	mov ah,09H
	mov dx,offset debug_spacer
	int 21H
	mov dl,debug_al
	mov ah,06H
	int 21H
	mov ah,09H
	mov dx,offset debug_spacer
	int 21H

	mov ah,06H
	mov dl,debug_bh
	int 21H
	mov ah,09H
	mov dx,offset debug_spacer
	int 21H
	mov dl,debug_bl
	mov ah,06H
	int 21H
	mov ah,09H
	mov dx,offset debug_spacer
	int 21H


	mov ah,06H
	mov dl,debug_ch
	int 21H
	mov ah,09H
	mov dx,offset debug_spacer
	int 21H
	mov dl,debug_cl
	mov ah,06H
	int 21H
	mov ah,09H
	mov dx,offset debug_spacer
	int 21H

	mov ah,06H
	mov dl,debug_dh
	int 21H
	mov ah,09H
	mov dx,offset debug_spacer
	int 21H
	mov dl,debug_dl
	mov ah,06H
	int 21H
	mov ah,09H
	mov dx,offset debug_newline
	int 21H

	pop dx
	pop ax
	pop bx
	pop ds
	ret
endp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;; Auxiliary Procedures ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; return in cx the length of the null terminated string pointed to
; by es:bx

strlen  proc
	push ax
	push ds
	push si
	mov ax,es
	mov ds,ax
	mov si,bx
	cld
	xor cx,cx
strlen_loop:
	lodsb
	cmp al,00H
	je strlen_out
	inc cx
	jmp strlen_loop
strlen_out:
	pop si
	pop ds
	pop ax
	ret
endp	

; return 0 in ax if string pointed to by es:bx is the same as
; the one pointed to by ds:dx   
; String pointed to by es:bx must 
; have length in cx  

strncmp proc
	push di
	push si
	inc cx
	mov di,bx
	mov si,dx
	cld
repe	cmpsb 
	mov ax,cx
	pop si
	pop di
	ret
endp

; print number 0-65k in ax as decimal to screen
; Assumes ds equals cs

convertee db 0,0,0,0,0,'$' ; buffer to hold digits
base dw 10             ; base for conversion

print_num proc

	push bx
	push dx
	mov bx,4       ; index of digit in convertee buffer
	mov base,10
convert_cycle:
	xor dx,dx
	div word ptr base
	add dx,48      ; convert remainder in dx to ascii
	mov convertee[bx],dl
	dec bx         ; point to next digit position
	cmp bx,0
	jne convert_cycle ; keep dividing until zero
	
	; now use dos service to print result
	; assume ds has been loaded properly for this segment
	mov dx, offset convertee
	mov ah,09H
	int 21H
	pop dx
	pop bx
	ret
endp

; Print version/specification for XMS
; The version is in ax,  specification in bx
; Both in BCD

XMSVERSION1 db 'XMS Version ','$'
XMSVERSION2 db ' (Specification ','$'
XMSVERSION3 db ')',0AH,0DH,'$'

print_version proc
	push ax
	push bx
	push dx

	; print start of line
	mov dx,offset XMSVERSION1
	mov ah,09H
	int 21H

; convert version code in ax to x.xx 
	mov base,16    ; repeated division by 16
	xor dx,dx      ; prepare to divide base into dx:ax
	div base
	add dx,48      ; convert remainder in dx to ascii
	mov convertee[4],dl
	xor dx,dx      ; repeat for other digits, right to left
	div base
	add dx,48
	mov convertee[3],dl
	mov convertee[2],46   ; ascii code for dot	
	xor dx,dx
	div base
	add dx,48
	mov convertee[1],dl
	mov convertee[0],32   ; ascii code for space
	
	; now use dos service to print result
	; assume ds has been loaded properly for this segment
	mov dx, offset convertee
	mov ah,09H
	int 21H

	; set up to print specifiaction
	mov dx,offset XMSVERSION2
	mov ah,09H
	int 21H

; Now repeat conversion with bx in place of ax
	mov ax,bx
	mov base,16    ; repeated division by 16
	xor dx,dx      ; prepare to divide base into dx:ax
	div base
	add dx,48      ; convert remainder in dx to ascii
	mov convertee[4],dl
	xor dx,dx      ; repeat for other digits, right to left
	div base
	add dx,48
	mov convertee[3],dl
	mov convertee[2],46   ; ascii code for dot	
	xor dx,dx
	div base
	add dx,48
	mov convertee[1],dl
	mov convertee[0],32   ; ascii code for space
	
	; now use dos service to print result
	; assume ds has been loaded properly for this segment
	mov dx, offset convertee
	mov ah,09H
	int 21H

	; and print end of line
	mov dx,offset XMSVERSION3
	mov ah,09H
	int 21H
	
	pop dx
	pop bx
	pop ax
	ret
endp

; Little debug routine to print the number now in al

MY_DEBUG db ' = decimal value of al register ','$'

print_al proc

	push ax
	push ds
	push dx
	add al,48
	mov byte ptr cs:MY_DEBUG,al
	mov ax,cs
	mov ds,ax
	mov dx,offset MY_DEBUG
	mov ah,09H
	int 21H
	pop dx
	pop ds
	pop ax
	ret

endp

; Search through device_names (mount points) array to see if anything
; matches the file name pointed to by ds:dx, either upper case or lower case
; return 0 in ax if found and index of file in bx

pre_uc_buf db 'C:\'  ; in case name is not fully qualified
uc_buf db 16 dup (0) ; buffer to hold ds:dx converted to upper case
stopper db '$'       ; terminator for debug 09H use

file_name_search proc


	push ds
	push es
	push cx
	push di
	push si

; copy string pointed to by ds:dx to uc_buf, converting lower case
; letters to upper case (subtract 20H) and leave everything else the same

	mov cx, 10H  ; copy 16 bytes
	mov si,dx
	mov di, offset uc_buf 
force_uc:
	lodsb       ; move byte at ds:si to al
	cmp al,7AH  ; 7aH = z
	ja copy_it
	cmp al,61H  ; 61H = a
	jb copy_it
	sub al, 20H ; convert to UC
copy_it:
	mov byte ptr cs:[di], al
	inc di
	loop force_uc
	
; make ds:di point to our upper case buffer or pre_uc_buf if not fully
; qualified
	mov ax,cs
	mov ds,ax
	mov dx, offset uc_buf
	mov di,dx
	mov al, byte ptr [di + 1]
	cmp al, 3AH   ; 3AH = :  does it start c: ?
	je skip_normalize
	mov dx, offset pre_uc_buf
skip_normalize:

;mov ah,09H
;int 21H

	mov ax,cs     
	mov es,ax
	mov bx, offset device_names; es:bx points at first mount point name
	xor di,di    ; file array index

; loop over mount point names to see if any match ds:dx name

file_name_loop:
	call strlen  ; return length of string es:bx in cx
	call strncmp ; compare file name with mount point name
	cmp ax,0000H
	je found_name ; and return 0 in ax  
	add bx,10H   ; else next mount point name
	add di, SIZE MY_FILE
	cmp byte ptr es:[bx],00H  ; is it NULL?
	jne file_name_loop
	mov ax,01H ; not found

found_name:
	mov bx,di
	pop si
 	pop di
	pop cx
	pop es
	pop ds
	ret
endp

; if all else fails in debugging, call this
; and it will get your attention
bomb proc
	bomb_loop:
	jmp bomb_loop
endp

ram:  ; below is frame buffer
_TEXT ends
end

;;;; stuff below not assembled 
;  stuff for time stamping files - mine it later

	; This section needs to be fixed. It is almost certainly wrong
	; right now and won't really do anything until we start keeping
	; a genuine ram-folder. When a file is closed, its directory
	; should be updated
	; Need to time stamp place-holder file. Get current date
	mov ah,2AH
	int 21H
		     ; this puts day of month in dl, month number 1,...12
	             ; in dh, and year number in cx (from 1980)
		     ; need to repackage this into one register	
	push bx      ; store because we need bx
	sub cx, 1980
 	xor bx,bx
	mov bl,dh 
	sal bl,1
	sal bl,1
	sal bl,1
	sal bl,1
	sal bl,1
	add bl,dl
	adc bh,cl    ; bx should now have date that 5701H wants in dx
	 
	mov ah,2CH
	int 21H      ; Get Time
		     ; military hour in ch, minutes in cl, seconds in dh
	mov dl,ch
	xor ch,ch
	sal cx,1
	sal cx,1
	sal cx,1
	sal cx,1
	sal cx,1
	shr dh,1
	add cl,dh
	xor dh,dh
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	sal dx,1
	add cx,dx    ; cx now has time format 5701H wants
	mov dx,bx    ; now dx had date 
	pop bx       ; handle back in bx
	mov ax,5701H
	int 21H      ; set file date and time


