; Truss.asm ( ASSEMBLER SOURCE CODE MODULE )
;       Installs a TSR that allocates a desired amount of memory 
;       and then remains resident until uninstalled. 
;       This module is called by a C-program via 
;       _setup(unsigned par) where par gives the number of paragraphs
;       of memory to allocate in total. The C-program can also call 
;       void uninstall(void), which frees memory and otherwise cleans up
;       after the camper. (This would normally be a second invocation of camp
;       later on.)
;
;       The routines _uninstall, and _setup below are non-resident.

;       The only practical way to invoke a tsr is via an interrupt, but which
;       interrupt to use? Many of them are used by the system and almost all of
;       them could be used by some other program that might be run while we
;       are resident. We settled on 01H, which is normally only used by
;       debuggers. (This can be changed -- except references to 01H in labels
;       and comments - by changing HOOK equ 01H below.) 
;
;	When it is wakened by 01H, _uninst first invokes the original
;       handler and waits for its return. In theory this should make the
;       presence of camp invisible to other programs, even ones that are
;       using int 01H. Then it merely
;       returns three addresses in registers:  It's own PSP segment address
;       in si (used as a handle to free memory), the stored address of the
;       01H handler that was present when camp was installed (so it can
;       be restored) in cx:dx, and the entry point of _unist itself in
;       es:bp. 

;	When it is called from main, uninstall first issues int 01 in order
;       retrieve the information _uninst reports. Then it makes sure that the
;       current  01H 
;       vector matches the address of _uninst in case some other handler might 
;       have been installed in the meantime. If everything is ok, it then
;       restores the old handler and asks DOS to free the memory the resident
;       routine has been camping on. 
;
;       Camp is an unusual TSR in that it doesn't use the keep program service
;       to terminate and stay resident. What it does is allocate a block
;       of memory of the desired size in the desired place and copy the
;       resident part of the code there. This phase of operation ends with a far
;       jump to the entry point _tsr of the copied code. The _tsr code at
;       the beginning of the resident part then turns around and de-allocates
;       the memory of the non-resident code before exiting to DOS in the
;       normal way. Actually, DOS doesn't seem to allow de-allocation of the
;       main program memory in the normal way. Instead, _tsr tricks DOS by
;       setting the owner of its own MCB to itself. Rather amazingly, 
;       that seems to work.

.MODEL tiny

_TEXT SEGMENT PUBLIC 'CODE'
ASSUME cs:_TEXT,ds:_TEXT,ss:_TEXT
PUBLIC _quiet,_setup,_uninstall, _strategy, _go_tsr,_LinkFlag
PUBLIC _set_umb_link, _restore_umb_link,_PSP_SEG,_driver_start,_driver_end 
PUBLIC _test1, _test1_start
EXTRN _code_bytes, _code_buffer, _header_info
HOOK equ 01H      ; interrupt for the tsr to listen for ... (bad English, that)

org 100h          ; end of psp
begin:            ; Program entry point label 
jmp near ptr crt  ; Jump around to beginning of C-runtime


;
; BEGINNING  OF RESIDENT ROUTINE
;       handle int 01h in order to send data back to
;       non-resident program for uninstall purposes.
;

; We get here by a far return from _setup when setup is complete

MY_NAME db 'RAM_FILE',0

_tsr:
		mov ax,cs     ; load ds so we can access data
		mov ds,ax

		mov dx, _PSP_SEG
		mov ah, 26H   ; create new PSP (needed ????)
		int 21H

; change ownership of our own memory arena

		mov ax,_PSP_SEG
		dec ax             ; point at arena header 1 par below psp
		mov es,ax
		mov ax,_PSP_SEG
		mov word ptr es:0001, ax  ; claim memory as our own
		mov ax, offset MY_NAME    ; and sign it for debug
		mov si,ax
		mov di,08H
		mov cx,08H
	rep	movsb
		

		xor al,al
		mov ah,4CH
		int 21H                        ; exit(0). This nukes the
                                               ; memory of non-resident stuff		

_resident:

; Resident data
_PSP_SEG dw ?          ; segment address of our PSP ( = _TEXT segment start)
_driver_start dw ?
_driver_end dw ?
_driver_unhook dw 41H  ; segment address of driver unhook routine
_test1_start dw ?
old01vseg dw 00H
old01voff dw 00H      ; address of old int 01H handler: for uninstall

goofy db 'Where is the library at, ... asshole?',0DH,0AH,'$'

_test1: ; here as an entry point to test far call to this resident code
	; void far_test1(void)
	push dx
	mov dx, offset goofy  ; Print a goofy message as a test
	mov ah,09H
	int 21H           ; Print our "copyright"
	pop dx
	retf
	
_uninst:              ; entry point of HOOK handler

;
; HOOK INT HANDLER
;
; This routine is invoked from truss (new incarnation) via HOOK'ed interrupt
; Its job is to return the address of the original handler
; in cx:dx 
; so that it can be restored in the _uninstall routine of camp.
; Also sends cs:_uninst in bp:es (the address of our entry point)
; so that it can be compared with the current interrupt vector
; for int 01H. If they don't agree, then some other routine
; has stolen the interrupt, and it's not safe to uninstall.
; Finally, it sends segment of PSP in si. This is actually not the PSP since
; the resident routine has none and was copied from its original location
; by _setup. It is the handle returned by the DOS allocate memory system
; call, and is used by _uninstall to free up memory.

	push ds
	mov ax,cs
	mov ds,ax    ; who knows where ds was pointing when the interrupt
                     ; occurred? We point it at our own code segment.
;
; CHAIN TO OLD HANDLER
;
	pushf                     ; The first 3 pushes get us back here when
	push cs                   ; there is rti from old handler
	mov ax, offset back_again
	push ax                   ; cs:offset back_again is return  
	push old01vseg            
	push old01voff           
	retf                      ; Going to old handler now. Bye!

back_again:                       ; hopefully  :-)


;  THE REMAINING CODE IS USED FOR UNINTSTALL

;  Kill the driver

	mov ax,offset _endtsr
	add ax,03H   ; ok, this is where the unhook routine is
	call ax


;
;  Now we are going to send back information in the registers, but we
;  wouldn't want to do this if some other program had taken over the
;  int  -- it wouldn't be expecting what we are going to send it.
;  So check to make sure that the current interrupt handler points
;  to us.  


	mov al,HOOK
	mov ah,35H  ; get address of current handler in es:bx
	int 21H
	cli         ; critical section
	mov ax,es
	cmp ax,_PSP_SEG
	jne no_can_do
	cmp bx,offset _uninst
	jne no_can_do
	jmp cont                  ; ok to continue with uninstall

no_can_do:
	sti
	int 21H     ; DOS
	pop ds      ; returning to whomever caused interrupt. A debugger
	iret        ; in the default case of 01H?

cont:   sti                  ; We're returning to our own _uninstall
	mov ax,_PSP_SEG       ; load up registers with all our info
	mov si,ax            ; so it can do the job
	mov ax,old01voff
	mov dx,ax
	mov ax,old01vseg
	mov cx,ax
	mov ax,cs
	mov bp,ax
	mov ax,offset _uninst
	mov es,ax
	pop ds
	iret

_driver_data:

; driver starts here
org 400H
_endtsr:            ; end of resident code 
;                   
;                   
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Non Resident Data
err_msg_1    db 'There is something wrong. Cannot install.0DH,0AH','$'
err_msg_2    db 'Cannot uninstall. No, or wrong interrupt vector present.',0DH,0AH,'$'
motto db '                         CHICKEN HAVEN SOFTWARE',0DH,0AH,'$' 
install db 'Truss 1.0 By Terry McConnell (2017) No rights reserved.',0DH,0AH,'$'
parlen db 16     ; for a div
_quiet db 0      ; flag to signal that operation should be quiet
par_req dw ?     ; size of memory footprint to camp on (desired)
_strategy dw ?   ; memory allocation strategy
strategy_old dw ?
_LinkFlag db ?   ; used for UMB access

;
; TRANSIENT ("SETUP") PART OF PROGRAM
; This gets called from main.c
;
;   void setup(unsigned);
;
_setup  proc    ; beginning of "transient" setup routine
	push bp ; calibrate stack frame (_cdecl assumed)
	mov bp,sp
;
;  UNLOAD ARGUMENT FROM THE STACK
;
	mov ax,[bp+4]       ; get argument passed by main (16 bit int)
	mov par_req,ax      ; store pars requested (1 word)

;  SET MEMORY ALLOCATION STRATEGY

	mov ax,5800H        ; Get current memory allocation strategy
	int 21H
	mov strategy_old,ax

	push bx
	mov ax,_strategy    ; Can be set with /a option (default = 00H)
	mov bx,ax
	mov ax,5801H        ; Set memory allocation strategy
	int 21H
	pop bx
;
;  SAVE SEGMENT ADDRESS OF OUR PSP (for uninstall)
;
	mov ah,51H        ; Get PSP address
	int 21H
	mov _PSP_SEG,bx    ;segment address of current PSP
			  ; Couldn't we just store ds? Probably

; Free the program's environment block. The tsr does not use it

        push es
	push bx
        mov ax, _PSP_SEG
        mov es,ax         ; store psp seg address in es
        mov ax, es:002CH  ;seg address of env block at offset 2CH in psp
        mov es,ax
        mov ah,49H        ; free allocated memory (environment block)
        int 21H
	pop bx
	pop es

;
;  SAVE ADDRESS OF OLD INTERRUPT 01H HANDLER (for uninstall)
;

	mov al,HOOK  ; save address of old handler
	mov ah,35H   ; using dos function 35h
	int 21H
	cli          ; critical section
	mov ax,es
	mov old01vseg,ax   ; save segment register of handler
	mov ax,bx
	mov old01voff,ax   ; save offset of handler
	sti

; ASK DOS FOR THE DESIRED MEMORY (PRETTY PLEASE !!!) 

	push bx
	mov bx,par_req  ; compute pars needed by resident program 
	mov ax,offset _endtsr  
	div parlen      ; divide by 16 = paragraph length in bytes
        xor ah,ah       ; clear remainder
        add ax,bx       ; 
        mov bx,ax       ; 
	mov ah, 48H     ; Allocate memory to pars in bx 
	int 21H
	mov _PSP_SEG,ax  ; Store handle of allocated memory
	mov ax, offset _test1 ; store address to call test routine
	mov _test1_start, ax

; RESTORE ORIGINAL MEMORY ALLOCATION STRATEGY

	mov bx, strategy_old
	mov ax, 5801H
	int 21H
	pop bx

; COPY RESIDENT AND DRIVER CODE TO ALLOCATED AREA
;
;   "Well, Doc, sometime when sumpfin' bad happen
;    it leave a little trace of itseln behine"  -- The Shining

	push cx
	push si
	push di
	push es

	mov ax, offset _endtsr    ; number of bytes to copy
	mov cx,ax
	cld                       ; copy up
	xor ax,ax
	mov si,ax                 ; copy from beginning of ds
	mov di,ax                 ; to beginning of es
	mov ax,_PSP_SEG 
	mov es,ax                 ; destination segment

rep	movsb                     ; copy 1 byte until cx = 0

	mov ax,offset _endtsr
	mov _driver_start,ax      ; store offset of driver code
	mov di,ax
	mov ax, _code_buffer      ; char * code_buffer
	mov si,ax                 ; continue copy from ds:si to es:di
	mov ax, _code_bytes       ; int code_bytes
	mov cx,ax
rep	movsb                     ; copy driver code too
	mov ax,di
	mov _driver_end,ax

	pop es
	pop di
	pop si
	pop cx

;
; INSTALL NEW HANDLER HOOK interrupt ( the resident part of this program, now
;                                relocated)

        mov dx, offset _uninst
	push ds
	mov ax,_PSP_SEG
	mov ds,ax               ; switch to ds of resident routine
        mov al,HOOK 
        mov ah,25H
        int 21H
	pop ds                 ; switch back to our own ds

;	mov dx, offset motto
;	mov ah,09H
;	int 21H           ;Print our motto
;	mov dx, offset install
;	mov ah,09H
;	int 21H           ; Print our "copyright"

gohome: 
	pop bp            ; restore bp
	ret

barf:       
	mov dx,offset err_msg_1
	mov ah,09H    ; display string DOS service
	int 21H       ; DOS
	pop bp         ; restore bp
	mov al,01H     ; return value signals type 1 error
	mov ah,4CH    ; return to DOS without installing TSR.
	int 21H         ; exit to DOS

_setup   endp

; void go_tsr(void)
; do a far return to the installed code
_go_tsr proc near

	push _PSP_SEG
	mov ax,offset _tsr
	push ax
	retf

_go_tsr endp

;
;  UNINSTALL ROUTINE
;

;  The installed tsr contains an int 01H handler whose purpose is to
;  return some stored data in the registers. The data returned are
;  si: segment address of the resident program's PSP
;  cx:dx address of the original int 01H handler
;  bp:es address of resident program's own int 01H handler
;  This routine refuses to uninstall the tsr if the
;  current handler address 
;  is different from our own interrupt vector
;  handler assignment-- that would mean some other handler had been
;  installed in the meantime. Assuming that one chains back to ours,
;  we had better not remove it. (Hint from Marc Johnson.)

_uninstall proc near
	push ds
	push si
	push es
	push bp
	push cx
	push dx

	mov al,HOOK  ; save address of current handler as es:bx
	mov ah,35H   ; using dos function 35h
	int 21H
	mov ax,es
	mov old01vseg,ax   ; save segment register of handler
	mov ax,bx
	mov old01voff,ax   ; save offset of handler

	;  OK, now we can uninstall our handler


; We must put everything back the way we found it: the old 
; int vector must be restored and the program memory must be freed.

	int HOOK   ; query resident routine for info
	mov ax,es  
	cmp ax, old01voff
        jne err2
	mov ax,bp
	cmp ax, old01vseg
        jne err2
	mov ax,cx       ; ds:dx wants address of handler to restore
	mov ds,ax
	mov al,HOOK
	mov ah,25H
	int 21H         ; restore original int handler
	mov ax,si
	mov es,ax     ; seg address of psp
	mov ah,49H
	int 21H       ; free allocated memory
byebye:
	pop dx
	pop cx
	pop bp
	pop es
	pop si
	pop ds
	ret           ; to main (where a success message is printed.)

err2:    mov dx, offset err_msg_2          ; print an error message and exit
	mov ah,09H
	int 21H
	mov al,01H    ; return value is error 1
	mov ah,4CH
	int 21H       ; return to DOS
_uninstall endp

_set_umb_link proc near

	push ax
	push bx
	mov ax,5802H     ; save current UMB link status 1 = linked 0 = not
	int 21H
	mov _LinkFlag,al

	mov bx, 0001H   ; enable UMB link
	mov ax,5803H
	int 21H
	pop bx
	pop ax
	ret             ; Did I forget this before? Yes. Results were ...
                        ; interesting

_set_umb_link endp

_restore_umb_link proc near

	push ax
	push bx
	xor bx,bx 
	mov bl,_LinkFlag
	mov ax,5803H
	int 21H
	pop bx
	pop ax
	ret

_restore_umb_link endp

crt:       ; C run-time stuff goes below
_TEXT ends

my_stack SEGMENT STACK  ; not really needed, but linker complains otherwise

dw 100H dup (?)

my_stack ends

end begin  ; sound's weird, doesn't it?
